@@ -12,17 +12,19 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from typing import cast, Optional, List, Iterable, Counter
+from typing import cast, Optional, List, Sequence, Union, Counter
 import json
 import time
 from ast import literal_eval
 from requests import post, get, put
 from pytket.backends import Backend, ResultHandle, CircuitStatus, StatusEnum
 from pytket.backends.backend import KwargTypes
+from pytket.backends.backendinfo import BackendInfo, fully_connected_backendinfo
 from pytket.backends.resulthandle import _ResultIdTuple
 from pytket.backends.backendresult import BackendResult
 from pytket.backends.backend_exceptions import CircuitNotRunError
 from pytket.circuit import Circuit, Qubit  # type: ignore
+from pytket.extensions.ionq._metadata import __extension_version__
 from pytket.passes import (  # type: ignore
     BasePass,
     SequencePass,
@@ -32,6 +34,7 @@ from pytket.passes import (  # type: ignore
     DecomposeBoxes,
     FlattenRegisters,
     RenameQubitsPass,
+    SimplifyInitial,
 )
 from pytket.predicates import (  # type: ignore
     GateSetPredicate,
@@ -42,9 +45,8 @@ from pytket.predicates import (  # type: ignore
     NoSymbolsPredicate,
     Predicate,
 )
-from pytket.routing import FullyConnected  # type: ignore
+from pytket.utils import prepare_circuit
 from pytket.utils.outcomearray import OutcomeArray
-from pytket.device import Device  # type: ignore
 from .ionq_convert import ionq_pass, ionq_gates, ionq_singleqs, tk_to_ionq
 from .config import IonQConfig
 
@@ -80,11 +82,12 @@ class IonQBackend(Backend):
     """
 
     _supports_counts = True
+    _supports_contextual_optimisation = True
     _persistent_handles = True
 
     def __init__(
         self,
-        device_name: Optional[str] = "qpu",
+        device_name: str = "qpu",
         api_key: Optional[str] = None,
         label: Optional[str] = "job",
     ):
@@ -101,7 +104,6 @@ class IonQBackend(Backend):
         """
         super().__init__()
         self._url = IONQ_JOBS_URL
-        self._device_name = device_name
         self._label = label
         config = IonQConfig.from_default_config_file()
 
@@ -111,14 +113,19 @@ class IonQBackend(Backend):
             raise IonQAuthenticationError()
 
         self._header = {"Authorization": f"apiKey {api_key}"}
-        self._max_n_qubits = IONQ_N_QUBITS
-        self._device = Device(FullyConnected(self._max_n_qubits))
-        self._qm = {Qubit(i): node for i, node in enumerate(self._device.nodes)}
+        self._backend_info = fully_connected_backendinfo(
+            type(self).__name__,
+            device_name,
+            __extension_version__,
+            IONQ_N_QUBITS,
+            ionq_gates,
+        )
+        self._qm = {Qubit(i): node for i, node in enumerate(self._backend_info.nodes)}
         self._MACHINE_DEBUG = False
 
     @property
-    def device(self) -> Optional[Device]:
-        return self._device
+    def backend_info(self) -> Optional[BackendInfo]:
+        return self._backend_info
 
     @property
     def required_predicates(self) -> List[Predicate]:
@@ -128,7 +135,7 @@ class IonQBackend(Backend):
             NoMidMeasurePredicate(),
             NoSymbolsPredicate(),
             GateSetPredicate(ionq_gates),
-            MaxNQubitsPredicate(self._max_n_qubits),
+            MaxNQubitsPredicate(self._backend_info.n_nodes),
         ]
         return preds
 
@@ -151,6 +158,7 @@ class IonQBackend(Backend):
                     FlattenRegisters(),
                     RenameQubitsPass(self._qm),
                     ionq_pass,
+                    SimplifyInitial(allow_classical=False, create_all_qubits=True),
                 ]
             )
         else:
@@ -165,18 +173,19 @@ class IonQBackend(Backend):
                         ionq_singleqs,
                         lambda a, b, c: Circuit(1).Rz(c, 0).Rx(b, 0).Rz(a, 0),
                     ),
+                    SimplifyInitial(allow_classical=False, create_all_qubits=True),
                 ]
             )
 
     @property
     def _result_id_type(self) -> _ResultIdTuple:
-        # job id, qubit no., measure permutation
-        return (str, int, str)
+        # job id, qubit no., measure permutation, ppcirc
+        return (str, int, str, str)
 
     def process_circuits(
         self,
-        circuits: Iterable[Circuit],
-        n_shots: Optional[int] = None,
+        circuits: Sequence[Circuit],
+        n_shots: Optional[Union[int, Sequence[int]]] = None,
         valid_check: bool = True,
         **kwargs: KwargTypes,
     ) -> List[ResultHandle]:
@@ -184,31 +193,54 @@ class IonQBackend(Backend):
         See :py:meth:`pytket.backends.Backend.process_circuits`.
         Supported kwargs: none.
         """
-        if n_shots is None or n_shots < 1:
-            raise ValueError("Parameter n_shots is required for this backend")
+        circuits = list(circuits)
+        n_shots_list: List[int] = []
+        if hasattr(n_shots, "__iter__"):
+            for n in cast(Sequence[Optional[int]], n_shots):
+                if n is None or n < 1:
+                    raise ValueError(
+                        "n_shots values are required for all circuits for this backend"
+                    )
+                n_shots_list.append(n)
+            if len(n_shots_list) != len(circuits):
+                raise ValueError("The length of n_shots and circuits must match")
+        else:
+            if n_shots is None:
+                raise ValueError("Parameter n_shots is required for this backend")
+            # convert n_shots to a list
+            n_shots_list = [cast(int, n_shots)] * len(circuits)
 
         if valid_check:
             self._check_all_circuits(circuits)
-        basebody = {
+
+        postprocess = kwargs.get("postprocess", False)
+
+        basebody: dict = {
             "lang": "json",
             "body": None,
-            "target": self._device_name,
-            "shots": n_shots,
+            "target": self._backend_info.device_name,
         }
         handles = []
-        for i, circ in enumerate(circuits):
+        for i, (circ, n_shots) in enumerate(zip(circuits, n_shots_list)):
             result = dict()
             bodycopy = basebody.copy()
-            (bodycopy["body"], meas) = tk_to_ionq(circ)  # type: ignore
+            if postprocess:
+                c0, ppcirc = prepare_circuit(circ, allow_classical=False)
+                ppcirc_rep = ppcirc.to_dict()
+            else:
+                c0, ppcirc_rep = circ, None
+            (bodycopy["body"], meas) = tk_to_ionq(c0)  # type: ignore
             if len(meas) == 0:
                 result["result"] = self.empty_result(circ, n_shots=n_shots)
             measures = json.dumps(meas)
             bodycopy["name"] = circ.name if circ.name else f"{self._label}_{i}"
+            bodycopy["shots"] = n_shots
             if self._MACHINE_DEBUG:
                 handle = ResultHandle(
                     _DEBUG_HANDLE_PREFIX + str(circ.n_qubits),
                     n_shots,
                     measures,
+                    json.dumps(ppcirc_rep),
                 )
                 handles.append(handle)
             else:
@@ -228,7 +260,7 @@ class IonQBackend(Backend):
 
                 # extract job ID from response
                 job_id = resp["id"]
-                handle = ResultHandle(job_id, n_shots, measures)
+                handle = ResultHandle(job_id, n_shots, measures, json.dumps(ppcirc_rep))
                 handles.append(handle)
         for handle in handles:
             self._cache[handle] = result
@@ -293,12 +325,18 @@ class IonQBackend(Backend):
                 sum_counts = sum(tket_counts.values())
                 diff = n_shots - sum_counts
                 tket_counts[max_array] += diff
+                ppcirc_rep = json.loads(cast(str, handle[3]))
+                ppcirc = (
+                    Circuit.from_dict(ppcirc_rep) if ppcirc_rep is not None else None
+                )
                 if handle in self._cache:
                     self._cache[handle].update(
-                        {"result": BackendResult(counts=tket_counts)}
+                        {"result": BackendResult(counts=tket_counts, ppcirc=ppcirc)}
                     )
                 else:
-                    self._cache[handle] = {"result": BackendResult(counts=tket_counts)}
+                    self._cache[handle] = {
+                        "result": BackendResult(counts=tket_counts, ppcirc=ppcirc)
+                    }
         return CircuitStatus(statenum)
 
     def get_result(self, handle: ResultHandle, **kwargs: KwargTypes) -> BackendResult:
